## Copyright (C) 2021 The Octave Project Developers
##
## This program is free software; you can redistribute it and/or modify it under
## the terms of the GNU General Public License as published by the Free Software
## Foundation; either version 3 of the License, or (at your option) any later
## version.
##
## This program is distributed in the hope that it will be useful, but WITHOUT
## ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
## FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
## details.
##
## You should have received a copy of the GNU General Public License along with
## this program; if not, see <http://www.gnu.org/licenses/>.

## -*- texinfo -*-
## @deftypefn {Function File} {@var{Q} =} block_pattern (@var{DELTA},@var{T})
## Assign each eigenvalue of upper triangular matrix @var{T} to a block.
## Eigenvalue T(i,i) is assigned to block @var{Q}(i).
## If eigenvalues T(i,i) and T(j,j) (i != j) are assigned to different blocks,
## then
##
## @verbatim
##  |T(i,i) - T(j,j)| > DELTA
## @end verbatim
##
## If eigenvalue T(i,i) is assigned to block @var{Q}(i) with more than one
## eigenvalue, then there is at least one other eigenvalue T(j,j) (i != j) 
## in block @var{Q}(i) with
##
## @verbatim
## |T(i,i) - T(j,j)| <= DELTA
## @end verbatim
## 
## Default value for @var{DELTA} is 0.1.
## 
## This function is an implementation of the block pattern
## algorithm in [1] (algorithm 9.5, page 227).
##
## References:
## 
## @itemize
## @item
## [1] @nospell{N.J. Higham}, @cite{Functions of Matrices}. SIAM, 2008.
## @end itemize
##
## @end deftypefn

function q = block_pattern (delta = 0.1, T)
  ## check input
  if (size (T)(1) != size (T)(2))
    error("block_pattern: input T must be square matrix");
  endif
  if(! istriu (T))
    error("block_pattern: input T must be upper triangular matrix");
  endif

  n = size (T)(1);
  q = -1 * ones (1,n); # initialize q
  p = 1;
  for i = 1:n # line 3 in algorithm 9.5   
    ## T(i,i) has not been assigned to a block => assign it to block p
    if (q(i) <= 0) # line 4 in algorithm 9.5
      q(i) = p;
      p++;
    endif
    b = q(i);
    for j = i+1:n # line 8 in algorithm 9.5
      if (q(j) != b) # line 10 in algorithm 9.5
        if (abs (T(i,i) - T(j,j)) <= delta) # line 11 in algorithm 9.5
          ## T(j,j) has not been assigned to a block.
          if (q(j) <= 0) # line 12 in algorithm 9.5
            q(j) = b;
          ## T(j,j) is member of block q(j) != b.
          else # line 14 in algorithm 9.5
            min_block = -1;
            max_block = -1;
            if (b < q(j))
              min_block = b;
              max_block = q(j);
            else
              min_block = q(j);
              max_block = b;
            endif

            ## Move elements from max_block to min_block.
            for k = 1:n # line 15 in algorithm 9.5
              if (q(k) == max_block)
                q(k) = min_block;
              endif
            endfor

            ## Reduce index of block q(k) by 1 if q(k) > max_block.
            for k = 1:n # line 16 in algorithm 9.5
              if (q(k) > max_block)
                q(k) = q(k) - 1;
              endif
            endfor
            p--; # line 17 in algorithm 9.5
          endif
        endif
      endif
    endfor
  endfor
endfunction

%!test
%! n = 6;
%! randn ("state", 42);
%! T = triu(randn(n),1);
%! T(1,1) = 1.02; T(3,3) = 1.00; 
%! T(2,2) = 3.04; T(4,4) = 3.02;
%! T(5,5) = 7; T(6,6) = 7.02;
%! q = block_pattern(0.1,T);
%! assert(q,[1,2,1,2,3,3]);

%!test
%! n = 8;
%! randn ("state", 42);
%! T = triu(randn(n),1);
%! T(1,1) = 1.02; T(3,3) = 1.00;
%! T(2,2) = 2.04; T(4,4) = 2.03;
%! T(5,5) = 1.01; T(6,6) = 2.01;
%! T(7,7) = 1.03; T(8,8) = 2.02;
%! q = block_pattern(0.1,T);
%! assert(q,[1,2,1,2,1,2,1,2]);
