## Copyright (C) 2021 The Octave Project Developers
##
## This program is free software; you can redistribute it and/or modify it under
## the terms of the GNU General Public License as published by the Free Software
## Foundation; either version 3 of the License, or (at your option) any later
## version.
##
## This program is distributed in the hope that it will be useful, but WITHOUT
## ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
## FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
## details.
##
## You should have received a copy of the GNU General Public License along with
## this program; if not, see <http://www.gnu.org/licenses/>.

## -*- texinfo -*- 
## @deftypefn {Function File} {@var{swaps} =} confluent_perm (@var{q})
##  Return a swapping strategy which converts @var{q} into a confluent
##  permutation @var{q'}.  According to [1] a confluent permutation of
##  n integers @var{q(1)}, @dots{}, @var{q(n)} is a permutation of @var{q}
##  such that any repeated integers are next to each other. 
##  @var{q} is an array of positive integers.  @var{swaps} is a Nx2 matrix
##  whose rows contain index pairs of q.  Row @var{swaps}(i,:) describes the
##  movement of element @var{q}(@var{swaps}(i,1)) to position @var{swaps}(i,2)
##  of @var{q}.
##
##  @example
##  @var{q} = (1,2,1,3,2,1);
##  [3,2;6,3;6,5] = confluent_perm(@var{q});
##  Apply first swap [3,2]
##  (1,2,1,3,2,1) -> (1,1,2,3,2,1)
##  Apply second swap [6,3]
##  (1,1,2,3,2,1) -> (1,1,1,2,3,2)
##  Apply third swap [6,5]
##  (1,1,1,2,3,2) -> (1,1,1,2,2,3)
##  @end example 
##
##  This function is an implementation of algorithm 4.2 in [1].
##
##  References:
##
##  @itemize
##  @item
##  [1] @nospell{Davies and Higham},
##  @cite{A SCHUR–PARLETT ALGORITHM FOR COMPUTING MATRIX FUNCTIONS},
##  SIAM J. MATRIX ANAL. APPL., 2003. 
##  @end itemize
##
## @end deftypefn

function retval = confluent_perm (q)
  G = [];
  ILST = [];
  IFST = [];
  max_cluster = max(q);
  phi = -ones(1,max_cluster) ;
  for cluster = 1:max_cluster
     phi(cluster) = length(find(q == cluster));
  endfor
  for cluster = 1:max_cluster
    avg_idx = sum(find(q == cluster)) / phi(cluster);
    G = [G; cluster avg_idx];
  endfor
  G = sortrows(G,[2]);
  beta = 1;
  for k = 1:max_cluster
    cluster = G(k,1);
    if (length(find(q(beta : beta + phi(cluster) - 1) != cluster)) > 0)
      f = find(q == cluster);
      u = beta:beta + phi(cluster) - 1;
      swaps = find(f - u != 0);
      if (length(swaps) > 0)
        ILST = [ILST;u(swaps)'];
        IFST = [IFST;f(swaps)'];
      endif
      v_tmp = beta:f(end);
      v = [];
      for l = 1:length (v_tmp)
        if(length (find (f == v_tmp(l) )) == 0)
          v = [v;v_tmp(l)];
        endif
      endfor
      q(u(end)+1:f(end)) = q(v);
      q(u) = ones(1,phi(cluster)) * cluster;
    endif
    beta = beta + phi(cluster);
  endfor
  res = [];
  res(:,1) = IFST;
  res(:,2) = ILST;
  retval = res;
endfunction

%!test
%! q = [1,2,1,3,2,1];
%! swaps = confluent_perm(q);
%! assert (swaps,[3,2;
%!                6,3;
%!                6,5]);

%!test
%! q = [1,2,4,1,2,2,3];
%! swaps = confluent_perm(q);
%! assert (swaps,[4,2;
%!                4,3]);




