## Copyright (C) 2021 The Octave Project Developers
##
## This program is free software; you can redistribute it and/or modify it under
## the terms of the GNU General Public License as published by the Free Software
## Foundation; either version 3 of the License, or (at your option) any later
## version.
##
## This program is distributed in the hope that it will be useful, but WITHOUT
## ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
## FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
## details.
##
## You should have received a copy of the GNU General Public License along with
## this program; if not, see <http://www.gnu.org/licenses/>.

## -*- texinfo -*-
## @deftypefn {Function File} {@var{retval} =} eval_atomic_block (@var{T}, @var{FHANDLE}, @var{DHANDLE})
## Compute @var{FHANDLE}(@var{T}) for upper triangular matrix @var{T}.
##
## The eigenvalues of @var{T} have to be @qcode{"close"}, see help text of
## function block_pattern for more information.
## 
## @var{FHANDLE} is a function handle which points to a scalar function
## whose Taylor series has an infinite radius of convergence.
##
## @var{DHANDLE} is a function handle with three
## input parameters. The call @code{DHANDLE(FHANDLE,k,x)} evaluates the
## k-th derivative of function handle FHANDLE at value x.
## 
## This function implements algorithm 9.4 from [1]. 
## 
## References:
##
## @itemize
## @item
## [1] @nospell{N.J. Higham}, @cite{Functions of Matrices}. SIAM, 2008.
## @end item
## @end itemize
##
## @end deftypefn

function retval = eval_atomic_block (T, FHANDLE, DHANDLE)
  if (size (T)(1) != size (T)(2))
    error("eval_atomic_block: T must be square matrix");
  endif
  if(! istriu (T))
    error("eval_atomic_block: Input T has to be upper triangular matrix");
  endif
  if(! is_function_handle (FHANDLE))
    error("eval_atomic_block: Input FHANDLE must be function handle");
  endif
  if(! is_function_handle (DHANDLE))
    error("eval_atomic_block: Input DHANDLE must be function handle");
  endif
  n = size(T)(1);
  Ident = eye(n);
  sigma = sum(diag(T)) / n;
  tol = eps;
  M = T - sigma * Ident;
  N = triu(T,1);
  e = ones(n,1);
  y = (Ident - arrayfun(@abs,N)) \ e;
  mu = norm(y,Inf);
  Fold = DHANDLE(FHANDLE, 0, sigma) * Ident;
  P = M;
  Fnew = [];
  ## TODO Throw exception, if s > 10^9
  for s = 1:10^9
    Fnew = Fold + DHANDLE(FHANDLE, s, sigma) * P;
    P = (P * M) / (s+1) ;
    if (norm (Fnew - Fold, "fro") <= tol * norm (Fnew, "fro"))
      tmp = [];
      for r = 0:n-1
         tmp = [tmp;
                max(abs(DHANDLE(FHANDLE, s + r, diag(T)))) / factorial(r) ];
      endfor
      tmp;
      DELTA = max (tmp);
      if (mu * DELTA * norm(P,"fro") <= tol * norm(Fnew,"fro"))
        break;
      endif
    endif
    Fold = Fnew;
  endfor
  retval = Fnew;
endfunction


%!test
%! T = [1.0 3; 0 1.1];
%! R1 = eval_atomic_block(T,@exp,@eval_derivative);
%! R2 = ones(2,2);
%! R2(1,1) = exp(1.0);
%! R2(1,2) = ((exp(1.1) -exp(1))/(1.1 - 1.0)) * 3.0;
%! R2(2,1) = 0.0;
%! R2(2,2) = exp(1.1);
%! assert (R1, R2, sqrt (eps));
