## Copyright (C) 2021 The Octave Project Developers
##
## This program is free software; you can redistribute it and/or modify it under
## the terms of the GNU General Public License as published by the Free Software
## Foundation; either version 3 of the License, or (at your option) any later
## version.
##
## This program is distributed in the hope that it will be useful, but WITHOUT
## ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
## FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
## details.
##
## You should have received a copy of the GNU General Public License along with
## this program; if not, see <http://www.gnu.org/licenses/>.

## -*- texinfo -*- 
## @deftypefn {Function File} {@var{retval} =} funm2 (@var{A}, @var{FHANDLE})
## @deftypefnx {Function File} {@var{retval} =} funm2 (@var{A}, @var{FHANDLE}, @var{DELTA})
## @deftypefnx {Function File} {@var{retval} =} funm2 (@var{A}, @var{FHANDLE}, @var{DELTA}, @var{DHANDLE})
## Compute matrix function @var{FHANDLE}(@var{A}).
## 
## @var{A} must be square matrix.  @var{FHANDLE} must be a function handle,
## which points to a scalar function whose Taylor series has an infinite
## radius of convergence.  funm2 supports the following values for
## @var{FHANDLE} by default. 
##  
## @itemize
## @item @@exp
## @item @@sin @@cos
## @item @@sinh @@cosh
## @end itemize
##
## Other values for @var{FHANDLE} are possible if argument @var{DHANDLE}
## is used in addition.
##
## If @code{A = U * T * U'} is a schur decomposition, argument @var{DELTA}
## controls the blocking of the eigenvalues of matrix T.  Default value for
## @var{DELTA} is 0.1.  See help text of function block_pattern for more
## information.
##
## @var{DHANDLE} is a function handle with three input arguments.
## The call @code{DHANDLE(FHANDLE,k,x)} evaluates the k-th derivative of
## function handle FHANDLE at value x.
##
## This function is an implementation of the Schur-Parlett algorithm
## in [1] (algorithm 9.6, page 228).
## 
## References:
##
## @itemize
## @item
## [1] @nospell{N.J. Higham}, @cite{Functions of Matrices}. SIAM, 2008.
## @item
## [2] @nospell{Davies and Higham},
## @cite{A Schur–Parlett algorithm for computing matrix functions},
## SIAM J. MATRIX ANAL. APPL., 2003. 
## @end itemize
##
## @end deftypefn

function retval = funm2 (A, FHANDLE, DELTA = 0.1, DHANDLE = @eval_derivative)
  default_functions = {"cos","sin","exp","sinh","cosh"}; 

  ## check input
  if (nargin != 2 && nargin != 3 && nargin != 4)
    print_usage ();
  endif
  if (size(A)(1) != size(A)(2))
    error( "funm2: A must be square matrix");
  endif
  if (! is_function_handle (FHANDLE))
    error("funm2: FHANDLE must be function handle");
  endif
  if (nargin <= 3)
    fname = (functions(FHANDLE)).function;
    if (! ismember (fname, default_functions))
      error (strcat ("funm2: Derivative of %s is not supported by default.",
            " Use argument DHANDLE in addition."), fname);
    endif
  endif
  if (nargin == 4)
    if (! is_function_handle (DHANDLE))
      error("funm2: DHANDLE must be function handle");
    endif
  endif
  
  ## schur decomposition  A = U * T * U'
  U = [];
  T = [];
  if (istriu (A))
    U = eye (size(A)(1));
    T = A; 
  else
    [U, T] = schur (A, "complex");
  endif 
  if (isdiag (T))
    retval = U * diag (FHANDLE (diag (T))) * U';
    return;
  endif

  ## blocking
  q = block_pattern (DELTA, T); 
  swaps = confluent_perm (q);

  ## apply swaps to q and T
  if (size (swaps)(1) > 0)
    for i = 1:size (swaps)(1) 
      [U, T] = trexc (U, T, swaps(i,1), swaps(i,2)); 
      if (swaps(i,1) > swaps(i,2))
        tmp = q(swaps(i,1));
        q(swaps(i,2)+1:swaps(i,1)) = q(swaps(i,2):swaps(i,1)-1);
        q(swaps(i,2)) = tmp;
      else
        tmp = q(swaps(i,1));
        q(swaps(i,1):swaps(i,2)-1) = q(swaps(i,1)+1:swaps(i,2));
        q(swaps(i,2)) = tmp;   
      endif
    endfor
  endif
  
  ## find start index and block length for each block in q
  bnum = max (q); # number of blocks in q
  bstart = - ones (1, bnum); # start indices of blocks in q
  bsize = - ones (1, bnum); # sizes of blocks in q
  curr_val = -1;
  bcounter = 1; 
  for i = 1:length(q)
    if (i == 1) # first block
      curr_val = q(i);
      bstart(bcounter) = i;
      bsize(bcounter) = 1;
    else
      if (q(i) != curr_val) # new block
        curr_val = q(i);
        bcounter++;
        bstart(bcounter) = i;
        bsize(bcounter) = 1;
      else  
        bsize(bcounter) = bsize(bcounter) + 1;
      endif
    endif
  endfor

  ## calculate F = FHANDLE(T)

  F = zeros(size (T)(1), size (T)(1));
  for j = 1:bnum
    ## index range (column/row) of j-th diagonal block
    jr = bstart(j):bstart(j) + bsize(j) - 1; 
    ## calculate j-th diagonal block of F
    F(jr,jr) = eval_atomic_block (T(jr,jr),FHANDLE,DHANDLE);
    if (j > 1) # calculate block column j of F 
      for i = j - 1: - 1:1
        ## row index range of block (i,j)
        ir = bstart(i):bstart(i) + bsize(i) - 1;
        ## calculate block (i,j) of F by solving sylvester equation (9.7) in [1]
        rhs = []; # right hand side of sylvester equation
        rhs = F(ir,ir) * T(ir,jr) - T(ir,jr) * F(jr,jr);
        if (i + 1 <= j - 1)
          for k = i + 1:j - 1
            ## index range (column/row) of blocks (i,k) and (k,j)
            kr = bstart(k):bstart(k) + bsize(k) - 1;
            rhs = rhs + F(ir,kr) * T(kr,jr) - T(ir,kr) * F(kr,jr);
          endfor
        endif
        F(ir,jr) = sylvester (T(ir,ir),- T(jr,jr),rhs);
      endfor
    endif
  endfor
  retval = U * F * U';
  return;
endfunction

## Test from [2]
%!test
%! A = gallery('invol',8)*pi;
%! B = funm2(A,@cos,0.1,@eval_cos_derivative);
%! assert(B,-eye(8),sqrt(eps));

## Test from [2]
%!test
%! A = gallery('invol',8)*pi;
%! B = funm2(A,@cos);
%! assert(B,-eye(8),sqrt(eps));

## Test 9) from John Burkardt, Matrix Exponential Tests,  
## https://people.sc.fsu.edu/~jburkardt/f77_src/test_matrix_exponential/test_matrix_exponential.html.
## Retrieved April 9, 2021.  
%!test
%! A = [ 1, 2, 2, 2; ...
%!        3, 1, 1, 2; ...
%!        3, 2, 1, 2; ...
%!        3, 3, 3, 1 ];
%! C = funm2(A,@exp); 
%! B = [ ...
%!         740.7038, 610.8500, 542.2743, 549.1753; ...
%!         731.2510, 603.5524, 535.0884, 542.2743; ...
%!         823.7630, 679.4257, 603.5524, 610.8500; ...
%!         998.4355, 823.7630, 731.2510, 740.7038 ];
%! assert(B,C,10^-4);

## Test 10) from John Burkardt, Matrix Exponential Tests,  
## https://people.sc.fsu.edu/~jburkardt/f77_src/test_matrix_exponential/test_matrix_exponential.html.
## Retrieved April 9, 2021.
%!test
%! A = [ 4, 2, 0; ...
%!       1, 4, 1; ...
%!       1, 1, 4 ];
%! C = funm2(A,@exp);
%! B = [ ...
%!         147.8666224463699, 183.7651386463682,  71.79703239999647; ...
%!         127.7810855231823, 183.7651386463682,  91.88256932318415; ...
%!         127.7810855231824, 163.6796017231806, 111.9681062463718 ];
%! assert(B,C,sqrt(eps));

%!test 
%!  A = zeros(9,9);
%!  A(1:3,1:3) = [3 1 1;0 3 1;0 0 3]; 
%!  A(4:6,4:6) = [2 1 1;0 2 1;0 0 2]; 
%!  A(7:9,7:9) = [1 1 1;0 1 1;0 0 1];
%!  A(1,8) = 1; A(1,9) = 1; 
%!  A(2,9) = 1;
%!  C = funm2(A,@exp);
%!  B = zeros(9,9);
%!  B(1:3,1:3) = [2.008553692318765e+01    2.008553692318760e+01   3.012830538478118e+01;  
%!                                    0    2.008553692318765e+01   2.008553692318760e+01;
%!                                    0                       0   2.008553692318765e+01];
%!
%!  B(4:6,4:6) = [7.389056098930649e+00   7.389056098930649e+00   1.108358414839597e+01;
%!                                    0   7.389056098930649e+00   7.389056098930649e+00;
%!                                    0                       0   7.389056098930649e+00];
%!
%! B(1,8) = 8.683627547364303e+00;   B(1,9) = 1.736725509472859e+01;
%! B(2,9) = 8.683627547364303e+00;
%!
%! B(7:9,7:9)  = [2.718281828459046e+00   2.718281828459046e+00   4.077422742688567e+00;
%!                                    0   2.718281828459046e+00   2.718281828459046e+00;
%!                                    0                       0   2.718281828459046e+00];
%! assert(B,C,sqrt(eps));
